__version__ = '0.0.6'
__folder_structure_version__ = 2

from .genbank_to_fasta import GenBankToFasta
from .import_genome import OgbImporter
from .metadata_schemas import organism_json_schema, genome_json_schema
from .reindex_assembly import reindex_assembly
from .rename_eggnog import rename_eggnog, EggnogFile
from .rename_gff import rename_gff, GffFile
from .rename_genbank import rename_genbank, GenBankFile
from .rename_custom_annotations import rename_custom_annotations, CustomAnnotationFile
from .rename_fasta import rename_fasta, FastaFile
from .download_ncbi_genome import download_ncbi_fna_gbk_gff, rename_ncbi_files, download_ncbi_genome
